@@ -1,23 +1,23 @@
-/* $Id: cut.c 5671 2016-02-23 12:07:58Z bens $ */
 /**************************************************************************
- *   cut.c                                                                *
+ *   cut.c  --  This file is part of GNU nano.                            *
  *                                                                        *
  *   Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007,  *
  *   2008, 2009, 2010, 2011, 2013, 2014 Free Software Foundation, Inc.    *
- *   This program is free software; you can redistribute it and/or modify *
- *   it under the terms of the GNU General Public License as published by *
- *   the Free Software Foundation; either version 3, or (at your option)  *
- *   any later version.                                                   *
+ *   Copyright (C) 2014 Mark Majeres                                      *
+ *   Copyright (C) 2016 Benno Schulenberg                                 *
  *                                                                        *
- *   This program is distributed in the hope that it will be useful, but  *
- *   WITHOUT ANY WARRANTY; without even the implied warranty of           *
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
- *   General Public License for more details.                             *
+ *   GNU nano is free software: you can redistribute it and/or modify     *
+ *   it under the terms of the GNU General Public License as published    *
+ *   by the Free Software Foundation, either version 3 of the License,    *
+ *   or (at your option) any later version.                               *
+ *                                                                        *
+ *   GNU nano is distributed in the hope that it will be useful,          *
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty          *
+ *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.              *
+ *   See the GNU General Public License for more details.                 *
  *                                                                        *
  *   You should have received a copy of the GNU General Public License    *
- *   along with this program; if not, write to the Free Software          *
- *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA            *
- *   02110-1301, USA.                                                     *
+ *   along with this program.  If not, see http://www.gnu.org/licenses/.  *
  *                                                                        *
  **************************************************************************/
 
@@ -205,7 +205,7 @@ void do_cut_text(
 #endif /* !NANO_TINY */
 	set_modified();
 
-    edit_refresh_needed = TRUE;
+    refresh_needed = TRUE;
 
 #ifndef DISABLE_COLOR
     reset_multis(openfile->current, FALSE);
@@ -270,6 +270,8 @@ void do_cut_till_eof(void)
 /* Copy text from the cutbuffer into the current filestruct. */
 void do_uncut_text(void)
 {
+    int was_lineno = openfile->current->lineno;
+
     assert(openfile->current != NULL && openfile->current->data != NULL);
 
     /* If the cutbuffer is empty, get out. */
@@ -288,6 +290,9 @@ void do_uncut_text(void)
     update_undo(PASTE);
 #endif
 
+    if (openfile->current->lineno - was_lineno < editwinrows)
+	focusing = FALSE;
+
     /* Set the current place we want to where the text from the
      * cutbuffer ends. */
     openfile->placewewant = xplustabs();
@@ -295,7 +300,7 @@ void do_uncut_text(void)
     /* Mark the file as modified. */
     set_modified();
 
-    edit_refresh_needed = TRUE;
+    refresh_needed = TRUE;
 
 #ifndef DISABLE_COLOR
     reset_multis(openfile->current, FALSE);
