@@ -1,22 +1,22 @@
 /**************************************************************************
- *   files.c                                                              *
+ *   files.c  --  This file is part of GNU nano.                          *
  *                                                                        *
  *   Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007,  *
  *   2008, 2009, 2010, 2011, 2013, 2014 Free Software Foundation, Inc.    *
- *   This program is free software; you can redistribute it and/or modify *
- *   it under the terms of the GNU General Public License as published by *
- *   the Free Software Foundation; either version 3, or (at your option)  *
- *   any later version.                                                   *
+ *   Copyright (C) 2015, 2016 Benno Schulenberg                           *
  *                                                                        *
- *   This program is distributed in the hope that it will be useful, but  *
- *   WITHOUT ANY WARRANTY; without even the implied warranty of           *
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
- *   General Public License for more details.                             *
+ *   GNU nano is free software: you can redistribute it and/or modify     *
+ *   it under the terms of the GNU General Public License as published    *
+ *   by the Free Software Foundation, either version 3 of the License,    *
+ *   or (at your option) any later version.                               *
+ *                                                                        *
+ *   GNU nano is distributed in the hope that it will be useful,          *
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty          *
+ *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.              *
+ *   See the GNU General Public License for more details.                 *
  *                                                                        *
  *   You should have received a copy of the GNU General Public License    *
- *   along with this program; if not, write to the Free Software          *
- *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA            *
- *   02110-1301, USA.                                                     *
+ *   along with this program.  If not, see http://www.gnu.org/licenses/.  *
  *                                                                        *
  **************************************************************************/
 
@@ -51,6 +51,8 @@ bool has_valid_path(const char *filename)
 	statusline(ALERT, _("Path '%s' is not a directory"), parentdir);
     else if (access(parentdir, X_OK) == -1)
 	statusline(ALERT, _("Path '%s' is not accessible"), parentdir);
+    else if (ISSET(LOCKING) && access(parentdir, W_OK) == -1)
+	statusline(MILD, _("Directory '%s' is not writable"), parentdir);
     else
 	validity = TRUE;
 
@@ -98,6 +100,7 @@ void make_new_buffer(void)
     openfile->mark_set = FALSE;
     openfile->mark_begin = NULL;
     openfile->mark_begin_x = 0;
+    openfile->kind_of_mark = SOFTMARK;
 
     openfile->fmt = NIX_FILE;
 
@@ -232,8 +235,7 @@ int write_lockfile(const char *lockfilename, const char *origfilename, bool modi
      * byte 0        - 0x62
      * byte 1        - 0x30
      * bytes 2-12    - program name which created the lock
-     * bytes 24,25   - little endian store of creator program's PID
-     *                 (b24 = 256^0 column, b25 = 256^1 column)
+     * bytes 24-27   - PID (little endian) of creator process
      * bytes 28-44   - username of who created the lock
      * bytes 68-100  - hostname of where the lock was created
      * bytes 108-876 - filename the lock is for
@@ -248,7 +250,9 @@ int write_lockfile(const char *lockfilename, const char *origfilename, bool modi
     lockdata[0] = 0x62;
     lockdata[1] = 0x30;
     lockdata[24] = mypid % 256;
-    lockdata[25] = mypid / 256;
+    lockdata[25] = (mypid / 256) % 256;
+    lockdata[26] = (mypid / (256 * 256)) % 256;
+    lockdata[27] = mypid / (256 * 256 * 256);
     snprintf(&lockdata[2], 11, "nano %s", VERSION);
     strncpy(&lockdata[28], mypwuid->pw_name, 16);
     strncpy(&lockdata[68], myhostname, 31);
@@ -318,7 +322,7 @@ int do_lockfile(const char *filename)
     if (stat(lockfilename, &fileinfo) != -1) {
 	ssize_t readtot = 0;
 	ssize_t readamt = 0;
-	char *lockbuf, *question, *postedname, *promptstr;
+	char *lockbuf, *question, *pidstring, *postedname, *promptstr;
 	int room, response;
 
 	if ((lockfd = open(lockfilename, O_RDONLY)) < 0) {
@@ -333,6 +337,8 @@ int do_lockfile(const char *filename)
 	    readtot += readamt;
 	} while (readamt > 0 && readtot < LOCKBUFSIZE);
 
+	close(lockfd);
+
 	if (readtot < 48) {
 	    statusline(MILD, _("Error reading lock file %s: "
 			"Not enough data read"), lockfilename);
@@ -341,18 +347,23 @@ int do_lockfile(const char *filename)
 	}
 
 	strncpy(lockprog, &lockbuf[2], 10);
-	lockpid = (unsigned char)lockbuf[25] * 256 + (unsigned char)lockbuf[24];
+	lockpid = (((unsigned char)lockbuf[27] * 256 + (unsigned char)lockbuf[26]) * 256 +
+			(unsigned char)lockbuf[25]) * 256 + (unsigned char)lockbuf[24];
 	strncpy(lockuser, &lockbuf[28], 16);
 	free(lockbuf);
 
+	pidstring = charalloc(11);
+	sprintf (pidstring, "%u", (unsigned int)lockpid);
+
 #ifdef DEBUG
 	fprintf(stderr, "lockpid = %d\n", lockpid);
 	fprintf(stderr, "program name which created this lock file should be %s\n", lockprog);
 	fprintf(stderr, "user which created this lock file should be %s\n", lockuser);
 #endif
 	/* TRANSLATORS: The second %s is the name of the user, the third that of the editor. */
-	question = _("File %s is being edited (by %s with %s, PID %d); continue?");
-	room = COLS - strlenpt(question) - strlenpt(lockuser) - strlenpt(lockprog) + 3;
+	question = _("File %s is being edited (by %s with %s, PID %s); continue?");
+	room = COLS - strlenpt(question) + 7 - strlenpt(lockuser) -
+				strlenpt(lockprog) - strlenpt(pidstring);
 	if (room < 4)
 	    postedname = mallocstrcpy(NULL, "_");
 	else if (room < strlenpt(filename)) {
@@ -365,11 +376,12 @@ int do_lockfile(const char *filename)
 	} else
 	    postedname = mallocstrcpy(NULL, filename);
 
-	/* Allow extra space for username (14), program name (8), PID (3),
+	/* Allow extra space for username (14), program name (8), PID (8),
 	 * and terminating \0 (1), minus the %s (2) for the file name. */
-	promptstr = charalloc(strlen(question) + 24 + strlen(postedname));
-	sprintf(promptstr, question, postedname, lockuser, lockprog, lockpid);
+	promptstr = charalloc(strlen(question) + 29 + strlen(postedname));
+	sprintf(promptstr, question, postedname, lockuser, lockprog, pidstring);
 	free(postedname);
+	free(pidstring);
 
 	response = do_yesno_prompt(FALSE, promptstr);
 	free(promptstr);
@@ -827,7 +839,7 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable, bool checkw
     if (len > 0) {
 #ifndef NANO_TINY
 	/* If file conversion isn't disabled and the last character in
-	 * this file is '\r', set format to Mac if we currently think
+	 * This file is '\r', set format to Mac if we currently think
 	 * the file is a *nix file, or to both DOS and Mac if we
 	 * currently think the file is a DOS file. */
 	if (buf[len - 1] == '\r' && !ISSET(NO_CONVERT) && format < 2)
@@ -1076,21 +1088,21 @@ void do_insertfile(
     while (TRUE) {
 #ifndef NANO_TINY
 	if (execute) {
-	    msg =
 #ifndef DISABLE_MULTIBUFFER
-		ISSET(MULTIBUFFER) ?
-		_("Command to execute in new buffer [from %s] ") :
+	    if (ISSET(MULTIBUFFER))
+		msg = _("Command to execute in new buffer [from %s] ");
+	    else
 #endif
-		_("Command to execute [from %s] ");
+		msg = _("Command to execute [from %s] ");
 	} else
 #endif /* NANO_TINY */
 	{
-	    msg =
 #ifndef DISABLE_MULTIBUFFER
-		ISSET(MULTIBUFFER) ?
-		_("File to insert into new buffer [from %s] ") :
+	    if (ISSET(MULTIBUFFER))
+		msg = _("File to insert into new buffer [from %s] ");
+	    else
 #endif
-		_("File to insert [from %s] ");
+		msg = _("File to insert [from %s] ");
 	}
 
 	present_path = mallocstrcpy(present_path, "./");
@@ -1108,8 +1120,8 @@ void do_insertfile(
 #endif
 		edit_refresh, msg,
 #ifndef DISABLE_OPERATINGDIR
-		operating_dir != NULL && strcmp(operating_dir,
-		".") != 0 ? operating_dir :
+		operating_dir != NULL && strcmp(operating_dir, ".") != 0 ?
+		operating_dir :
 #endif
 		"./");
 
@@ -1226,7 +1238,7 @@ void do_insertfile(
 	    {
 		/* Make sure the path to the file specified in answer is
 		 * tilde-expanded. */
-		answer = mallocstrassn(answer, real_dir_from_tilde(answer));
+		answer = free_and_assign(answer, real_dir_from_tilde(answer));
 
 		/* Save the file specified in answer in the current buffer. */
 		open_buffer(answer, TRUE);
@@ -1695,15 +1707,15 @@ int copy_file(FILE *inn, FILE *out)
  *
  * tmp means we are writing a temporary file in a secure fashion.  We
  * use it when spell checking or dumping the file on an error.  If
- * append is APPEND, it means we are appending instead of overwriting.
- * If append is PREPEND, it means we are prepending instead of
+ * method is APPEND, it means we are appending instead of overwriting.
+ * If method is PREPEND, it means we are prepending instead of
  * overwriting.  If nonamechange is TRUE, we don't change the current
- * filename.  nonamechange is ignored if tmp is FALSE, we're appending,
- * or we're prepending.
+ * filename.  nonamechange is irrelevant when appending or prepending,
+ * or when writing a temporary file.
  *
  * Return TRUE on success or FALSE on error. */
-bool write_file(const char *name, FILE *f_open, bool tmp, append_type
-	append, bool nonamechange)
+bool write_file(const char *name, FILE *f_open, bool tmp,
+	kind_of_writing_type method, bool nonamechange)
 {
     bool retval = FALSE;
 	/* Instead of returning in this function, you should always
@@ -1769,7 +1781,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
      * only if the file has not been modified by someone else since nano
      * opened it. */
     if (ISSET(BACKUP_FILE) && !tmp && realexists && openfile->current_stat &&
-		(append != OVERWRITE || openfile->mark_set ||
+		(method != OVERWRITE || openfile->mark_set ||
 		openfile->current_stat->st_mtime == st.st_mtime)) {
 	int backup_fd;
 	FILE *backup_file;
@@ -1948,7 +1960,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
     }
 
     /* If we're prepending, copy the file to a temp file. */
-    if (append == PREPEND) {
+    if (method == PREPEND) {
 	int fd_source;
 	FILE *f_source = NULL;
 
@@ -1965,7 +1977,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	tempname = safe_tempfile(&f);
 
 	if (tempname == NULL) {
-	    statusline(HUSH, _("Error writing temp file: %s"),
+	    statusline(ALERT, _("Error writing temp file: %s"),
 			strerror(errno));
 	    goto cleanup_and_exit;
 	}
@@ -1987,7 +1999,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	}
 
 	if (f_source == NULL || copy_file(f_source, f) != 0) {
-	    statusline(HUSH, _("Error writing %s: %s"), tempname,
+	    statusline(ALERT, _("Error writing temp file: %s"),
 			strerror(errno));
 	    unlink(tempname);
 	    goto cleanup_and_exit;
@@ -1997,7 +2009,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
     if (f_open == NULL) {
 	/* Now open the file in place.  Use O_EXCL if tmp is TRUE.  This
 	 * is copied from joe, because wiggy says so *shrug*. */
-	fd = open(realname, O_WRONLY | O_CREAT | ((append == APPEND) ?
+	fd = open(realname, O_WRONLY | O_CREAT | ((method == APPEND) ?
 		O_APPEND : (tmp ? O_EXCL : O_TRUNC)), S_IRUSR |
 		S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
 
@@ -2006,17 +2018,17 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 
 	/* If we couldn't open the file, give up. */
 	if (fd == -1) {
-	    statusline(HUSH, _("Error writing %s: %s"), realname,
+	    statusline(ALERT, _("Error writing %s: %s"), realname,
 			strerror(errno));
 	    if (tempname != NULL)
 		unlink(tempname);
 	    goto cleanup_and_exit;
 	}
 
-	f = fdopen(fd, (append == APPEND) ? "ab" : "wb");
+	f = fdopen(fd, (method == APPEND) ? "ab" : "wb");
 
 	if (f == NULL) {
-	    statusline(HUSH, _("Error writing %s: %s"), realname,
+	    statusline(ALERT, _("Error writing %s: %s"), realname,
 			strerror(errno));
 	    close(fd);
 	    goto cleanup_and_exit;
@@ -2040,7 +2052,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	unsunder(fileptr->data, data_len);
 
 	if (size < data_len) {
-	    statusline(HUSH, _("Error writing %s: %s"), realname,
+	    statusline(ALERT, _("Error writing %s: %s"), realname,
 			strerror(errno));
 	    fclose(f);
 	    goto cleanup_and_exit;
@@ -2057,7 +2069,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 #ifndef NANO_TINY
 	    if (openfile->fmt == DOS_FILE || openfile->fmt == MAC_FILE) {
 		if (putc('\r', f) == EOF) {
-		    statusline(HUSH, _("Error writing %s: %s"), realname,
+		    statusline(ALERT, _("Error writing %s: %s"), realname,
 				strerror(errno));
 		    fclose(f);
 		    goto cleanup_and_exit;
@@ -2067,7 +2079,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	    if (openfile->fmt != MAC_FILE)
 #endif
 		if (putc('\n', f) == EOF) {
-		    statusline(HUSH, _("Error writing %s: %s"), realname,
+		    statusline(ALERT, _("Error writing %s: %s"), realname,
 				strerror(errno));
 		    fclose(f);
 		    goto cleanup_and_exit;
@@ -2079,7 +2091,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
     }
 
     /* If we're prepending, open the temp file, and append it to f. */
-    if (append == PREPEND) {
+    if (method == PREPEND) {
 	int fd_source;
 	FILE *f_source = NULL;
 
@@ -2099,23 +2111,23 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	}
 
 	if (copy_file(f_source, f) == -1) {
-	    statusline(HUSH, _("Error writing %s: %s"), realname,
+	    statusline(ALERT, _("Error writing %s: %s"), realname,
 			strerror(errno));
 	    goto cleanup_and_exit;
 	}
 
 	unlink(tempname);
     } else if (fclose(f) != 0) {
-	statusline(HUSH, _("Error writing %s: %s"), realname,
+	statusline(ALERT, _("Error writing %s: %s"), realname,
 			strerror(errno));
 	goto cleanup_and_exit;
     }
 
-    if (!tmp && append == OVERWRITE) {
+    if (method == OVERWRITE && !tmp) {
 	/* If we must set the filename, and it changed, adjust things. */
 	if (!nonamechange && strcmp(openfile->filename, realname) != 0) {
 #ifndef DISABLE_COLOR
-	    char *syntaxname = openfile->syntax ? openfile->syntax->name : "";
+	    char *oldname = openfile->syntax ? openfile->syntax->name : "";
 	    filestruct *line = openfile->fileage;
 #endif
 	    openfile->filename = mallocstrcpy(openfile->filename, realname);
@@ -2125,19 +2137,17 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 	    color_update();
 	    color_init();
 
+	    char *newname = openfile->syntax ? openfile->syntax->name : "";
+
 	    /* If the syntax changed, discard and recompute the multidata. */
-	    if (openfile->syntax &&
-			strcmp(syntaxname, openfile->syntax->name) != 0) {
+	    if (strcmp(oldname, newname) != 0) {
 		for (; line != NULL; line = line->next) {
 		    free(line->multidata);
 		    line->multidata = NULL;
 		}
 		precalc_multicolorinfo();
+		refresh_needed = TRUE;
 	    }
-
-	    /* If color syntaxes are available and turned on, refresh. */
-	    if (openfile->colorstrings != NULL && !ISSET(NO_COLOR_SYNTAX))
-		edit_refresh();
 #endif
 	}
 
@@ -2166,7 +2176,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp, append_type
 /* Write a marked selection from a file out to disk.  Return TRUE on
  * success or FALSE on error. */
 bool write_marked_file(const char *name, FILE *f_open, bool tmp,
-	append_type append)
+	kind_of_writing_type method)
 {
     bool retval;
     bool old_modified = openfile->modified;
@@ -2190,7 +2200,7 @@ bool write_marked_file(const char *name, FILE *f_open, bool tmp,
 	added_magicline = TRUE;
     }
 
-    retval = write_file(name, f_open, tmp, append, TRUE);
+    retval = write_file(name, f_open, tmp, method, TRUE);
 
     /* If we added a magicline, remove it now. */
     if (added_magicline)
@@ -2216,7 +2226,7 @@ int do_writeout(bool exiting)
 {
     int i;
     bool result = FALSE;
-    append_type append = OVERWRITE;
+    kind_of_writing_type method = OVERWRITE;
     char *given;
 	/* The filename we offer, or what the user typed so far. */
     bool maychange = (openfile->filename[0] == '\0');
@@ -2252,13 +2262,13 @@ int do_writeout(bool exiting)
 	 * it allows reading from or writing to files not specified on
 	 * the command line. */
 	if (openfile->mark_set && !exiting && !ISSET(RESTRICTED))
-	    msg = (append == PREPEND) ? _("Prepend Selection to File") :
-			(append == APPEND) ? _("Append Selection to File") :
+	    msg = (method == PREPEND) ? _("Prepend Selection to File") :
+			(method == APPEND) ? _("Append Selection to File") :
 			_("Write Selection to File");
 	else
 #endif /* !NANO_TINY */
-	    msg = (append == PREPEND) ? _("File Name to Prepend to") :
-			(append == APPEND) ? _("File Name to Append to") :
+	    msg = (method == PREPEND) ? _("File Name to Prepend to") :
+			(method == APPEND) ? _("File Name to Append to") :
 			_("File Name to Write");
 
 	present_path = mallocstrcpy(present_path, "./");
@@ -2334,10 +2344,10 @@ int do_writeout(bool exiting)
 	    } else
 #endif /* !NANO_TINY */
 	    if (func == prepend_void) {
-		append = (append == PREPEND) ? OVERWRITE : PREPEND;
+		method = (method == PREPEND) ? OVERWRITE : PREPEND;
 		continue;
 	    } else if (func == append_void) {
-		append = (append == APPEND) ? OVERWRITE : APPEND;
+		method = (method == APPEND) ? OVERWRITE : APPEND;
 		continue;
 	    } else if (func == do_help_void) {
 		continue;
@@ -2362,7 +2372,7 @@ int do_writeout(bool exiting)
 	    }
 #endif
 
-	    if (append == OVERWRITE) {
+	    if (method == OVERWRITE) {
 		size_t answer_len = strlen(answer);
 		bool name_exists, do_warning;
 		char *full_answer, *full_filename;
@@ -2451,10 +2461,10 @@ int do_writeout(bool exiting)
 	     * writing to files not specified on the command line. */
 #ifndef NANO_TINY
 	    if (openfile->mark_set && !exiting && !ISSET(RESTRICTED))
-		result = write_marked_file(answer, NULL, FALSE, append);
+		result = write_marked_file(answer, NULL, FALSE, method);
 	    else
 #endif
-		result = write_file(answer, NULL, FALSE, append, FALSE);
+		result = write_file(answer, NULL, FALSE, method, FALSE);
 
 	    break;
 	}
